import re
import time
from typing import Dict, List, Union

from broadcast_service import broadcast_service

from promptulate.llms.base import BaseLLM
from promptulate.llms.openai import ChatOpenAI
from promptulate.tools.arxiv.api_wrapper import ArxivAPIWrapper
from promptulate.tools.base import Tool
from promptulate.utils.core_utils import listdict_to_string, record_time


class ArxivQueryTool(Tool):
    name: str = "arxiv-query"
    description: str = (
        "A query tool around Arxiv.org "
        "Useful for when you need to answer questions about Physics, Mathematics, "
        "Computer Science, Quantitative Biology, Quantitative Finance, Statistics, "
        "Electrical Engineering, and Economics "
        "from scientific articles on arxiv.org. "
        "Input should be a search query."
    )
    api_wrapper: ArxivAPIWrapper = ArxivAPIWrapper()

    def _run(self, query: str, **kwargs) -> Union[str, List[Dict]]:
        """Arxiv query tool

        Args:
            query: the paper keyword or arxiv id
            **kwargs:
                return_type(Optional(str)):  return string default. If you want to
                return List[Dict] type data, you can set 'return_type'='original'.
                Moreover, you can pass the arguments of ArxivAPIWrapper

        Returns:
            String type default or List[Dict] arxiv query result.
        """
        kwargs.update({"from_callback": self.name})
        if re.match(r"\d{4}\.\d{5}(v\d+)?", query):
            result = self.api_wrapper.query(id_list=[query], **kwargs)
        else:
            result = self.api_wrapper.query(query, **kwargs)
        if "return_type" in kwargs and kwargs["return_type"] == "original":
            return result
        return listdict_to_string(result)


def _init_arxiv_reference_tool_llm():
    preset = "You are an Arxiv assistant whose task is to assist users in providing advice on academic papers. Your output can only follow the user's instructions, otherwise you will be punished."  # noqa: E501
    return ChatOpenAI(default_system_prompt=preset, temperature=0)


class ArxivReferenceTool(Tool):
    name: str = "arxiv-reference"
    description: str = (
        "Use this tool to find search related to the field."
        "Your input is a arxiv keyword query."
    )
    max_reference_num = 3
    reference_string: str = ""
    reference_counter: int = 0

    def __init__(self, llm: BaseLLM = None, **kwargs):
        self.llm: BaseLLM = llm or _init_arxiv_reference_tool_llm()
        super().__init__(**kwargs)

    @record_time()
    def _run(self, query: str, *args, **kwargs) -> Union[str, List[Dict]]:
        """
        Input arxiv paper information and return paper references.
        Args:
            query(str): arxiv paper information
            *args: Nothing
            **kwargs:
                return_type(Optional[str]): return string default. If you want to return
                List[Dict] type data, you can set 'return_type'='original.'
        Returns:
            String type or List[Dict] reference information
        """

        @broadcast_service.on_listen("ArxivReferenceTool.get_relevant_paper_info")
        @record_time()
        def get_relevant_paper_info(keyword: str):
            """return paper related information from keyword"""
            arxiv_query_tool = ArxivQueryTool()
            specified_fields = ["entry_id", "title"]
            queryset = arxiv_query_tool.run(
                keyword, num_results=6, specified_fields=specified_fields
            )
            self.reference_string += queryset
            self.reference_counter += 1

        def analyze_query_string(query_string: str) -> List[str]:
            """Analyze `[query]: keyword1, keyword2, keyword3` type data to return
            keywords list."""
            assert "[query]" in query_string
            query_string = query_string.split(":")[1]
            return query_string.split(",")

        def analyze_reference_string(reference_string: str) -> List[Dict]:
            pattern = r"\[(\d+)\]\s+(.+?),\s+(http://.+?);"
            references = []
            for match in re.finditer(pattern, reference_string):
                references.append({"title": match.group(2), "url": match.group(3)})
            return references

        self.reference_counter = 0
        prompt = (
            f"现在你需要根据其研究的具体内容，列出至少{self.max_reference_num}篇参考文献，你可以使用arxiv进行查询，你需要给我提供3个arxiv查询关键词，我将使用"
            f"arxiv进行查询，你需要根据我返回的结果选取最符合当前研究的{self.max_reference_num}篇参考文献。"
            f"用户输入:{query}"
            "你的输出必须是三个查询词\n，如 [query]: keyword1, keyword2, keyword3\n"
        )
        keywords = analyze_query_string(self.llm(prompt))
        for keyword in keywords:
            broadcast_service.broadcast(
                "ArxivReferenceTool.get_relevant_paper_info", keyword
            )

        while self.reference_counter < 3:
            time.sleep(0.2)

        prompt = (
            "Now you need to return the most appropriate 5 references \
            based on the papers given below\n"
            f"```{self.reference_string}```\n"
            "Your output format must be as follows:\nReferences:\n[1] [title1](url1);\n[2] [title2](url2);\n[3] [title3](url3);\n"  # noqa
        )
        # TODO If there is a problem with the returned format and an error is reported
        # then ask LLM to format the data
        result = self.llm(prompt)
        if "return_type" in kwargs and kwargs["return_type"] == "original":
            return analyze_reference_string(result)
        return result


def _init_arxiv_summary_tool_llm():
    preset = """You are an Arxiv assistant whose task is to assist users in providing advice and assistance on academic papers. Your output can only follow the user's instructions, otherwise you will be punished."""  # Noqa
    return ChatOpenAI(default_system_prompt=preset, temperature=0)


class ArxivSummaryTool(Tool):
    name: str = "arxiv-summary"
    description: str = (
        "A summary tool that can be used to obtain a paper summary, listing "
        "key insights and lessons learned in the paper, and references in the paper"
        "Your input is a arxiv keyword query."
    )
    summary_string: str = ""
    summary_counter: int = 0

    def __init__(self, llm: BaseLLM = None, **kwargs):
        self.llm: BaseLLM = llm or _init_arxiv_summary_tool_llm()
        self.api_wrapper: ArxivAPIWrapper = ArxivAPIWrapper()
        super().__init__(**kwargs)

    def _run(self, query: str, **kwargs) -> str:
        """An arxiv paper summary tool that passes in the article name (or arxiv id) and
        returns summary results.

        Args:
            query: the keyword you want to query
            **kwargs:
                You can pass the arguments of ArxivAPIWrapper

        Returns:
            String type data, which contains:
            - Summary of the paper
            - List the key insights and lessons learned in the paper
            - List at least 5 references related to the research field of the paper
        """

        @broadcast_service.on_listen("ArxivSummaryTool.run.get_opinion")
        def get_opinion():
            prompt = (
                f"Based on the abstract of the paper below, please list the key insights from the paper and the lessons learned derived from it. Your output should be provided in bullet points ```{paper_summary}```"  # noqa
                "Please use the following output format:\nKey insights:\n{Key insights in bullet points}\nLessons learned:\n{Lessons learned in bullet points},with each point separated by a hyphen `-`"  # noqa
            )
            opinion = self.llm(prompt)
            self.summary_string += opinion + "\n"
            self.summary_counter += 1

        @broadcast_service.on_listen("ArxivSummaryTool.run.get_references")
        def get_references():
            arxiv_referencer_tool = ArxivReferenceTool()
            references = arxiv_referencer_tool.run(paper_summary)
            self.summary_string += references + "\n\n"
            self.summary_counter += 1

        @broadcast_service.on_listen("ArxivSummaryTool.run.get_advice")
        def get_advice():
            prompt = (
                f"Based on the abstract of the paper below, please provide 3-5 suggestions related to its topics or potential future research directions. Your output should be provided in bullet points  ```{paper_summary}```"  # noqa
                "Please use the following output format:\nRelated suggestions:\n{related suggestions in bullet points}, with each point separated by a hyphen `-`"  # noqa
            )
            opinion = self.llm(prompt)
            self.summary_string += opinion + "\n"
            self.summary_counter += 1

        self.summary_counter = 0
        if re.match(r"\d{4}\.\d{5}(v\d+)?", query):
            paper_info = self.api_wrapper.query(
                id_list=[query], num_results=1, specified_fields=["title", "summary"]
            )
        else:
            paper_info = self.api_wrapper.query(
                query, num_results=1, specified_fields=["title", "summary"]
            )
            if len(paper_info) == 0:
                return "Could not find relevant arxiv article."
        paper_summary = listdict_to_string(paper_info, item_suffix="\n")
        self.summary_string = paper_summary

        broadcast_service.publish("ArxivSummaryTool.run.get_references")
        time.sleep(0.01)
        broadcast_service.publish("ArxivSummaryTool.run.get_opinion")
        time.sleep(0.01)
        broadcast_service.publish("ArxivSummaryTool.run.get_advice")

        while self.summary_counter < 3:
            time.sleep(0.1)

        return self.summary_string



from enum import Enum
from typing import Callable, Dict, List, Optional, TypedDict

from typing_extensions import NotRequired

from promptulate import uacp
from promptulate.agents.assistant_agent import operations
from promptulate.agents.assistant_agent.schema import Plan
from promptulate.agents.base import BaseAgent
from promptulate.agents.tool_agent import ToolAgent
from promptulate.agents.tool_agent.agent import ActionResponse
from promptulate.hook import Hook, HookTable
from promptulate.llms.base import BaseLLM
from promptulate.tools.base import ToolTypes
from promptulate.tools.manager import ToolManager
from promptulate.utils.logger import logger


class StepTypes(str, Enum):
    PLAN = "plan"
    EXECUTE = "execute"
    REVISE = "revise"


class AdditionalProperties(TypedDict):
    current_plan: dict
    past_steps: NotRequired[str]


class AssistantAgent(BaseAgent):
    """
    An agent who can plan, execute, and revise tasks.
    """

    def __init__(
        self,
        *,
        llm: BaseLLM,
        tools: Optional[List[ToolTypes]] = None,
        max_iterations: Optional[int] = 20,
        **kwargs,
    ):
        super().__init__(agent_type="Assistant Agent", **kwargs)

        self.llm = llm
        self.tool_manager = ToolManager(tools=tools if tools else [])
        self.tool_agent = ToolAgent(
            llm=llm, tool_manager=self.tool_manager, _from="agent"
        )
        self.uacp_agent = uacp.Agent(
            self.task_handler, self.step_handler, self.result_handler
        )
        self.current_task_id: Optional[str] = None
        self.max_iterations: int = max_iterations

        logger.info("Assistant Agent initialized.")

    def _run(
        self, instruction: str, additional_input: dict = None, *args, **kwargs
    ) -> str:
        additional_input = additional_input or {}

        result: str = self.uacp_agent.run(
            input=instruction, additional_input=additional_input
        )
        logger.info(f"Assistant Agent response: {result}")
        return result

    def get_llm(self) -> BaseLLM:
        return self.llm

    @property
    def current_task(self) -> Optional[uacp.Task]:
        """Get the current task, return None if no task is running."""
        if self.current_task_id is None:
            return None

        return self.uacp_agent.db.get_task(self.current_task_id)

    @property
    def current_plan(self) -> Optional[Plan]:
        """Get the current plan. Every step has a current plan stored in
        additional_properties."""
        if self.current_task is None:
            return None

        _: dict = self.current_task.steps[-1].additional_properties.get("current_plan")
        return Plan.parse_obj(_)

    @property
    def execution_steps(self) -> List[uacp.Step]:
        """Get the execution steps from the current task."""
        if self.current_task is None:
            return []

        return [s for s in self.current_task.steps if s.name == StepTypes.EXECUTE]

    def plan(self, step: uacp.Step) -> uacp.Step:
        """Plan the task and create the next step: execute.

        Args:
            step(uacp.Step): The current step.

        Returns:
            uacp.Step: The updated plan step.
        """
        logger.info("[Assistant Agent] Planning now.")

        current_plan: Plan = operations.plan(self.llm, step.input)
        Hook.call_hook(HookTable.ON_AGENT_PLAN, self, plan=current_plan.json())

        # create next step: execute
        task = self.uacp_agent.db.get_task(step.task_id)

        self.uacp_agent.db.create_step(
            task_id=task.task_id,
            name=StepTypes.EXECUTE,
            input=current_plan.tasks[0].description,
            additional_properties=AdditionalProperties(
                current_plan=current_plan.dict()
            ),
        )
        step.output = current_plan.json()

        return step

    def execute(self, step: uacp.Step) -> uacp.Step:
        """Execute the plan and create the next step: revise.

        Args:
            step(uacp.Step): The current step.

        Returns:
            uacp.Step: The updated execute step.
        """
        logger.info("[Assistant Agent] Executing step now.")

        resp: ActionResponse = operations.execute(self.tool_agent, step.input)

        if self.current_plan.get_next_task() is None:
            step.is_last = True
            step.output = resp["action_parameters"]["content"]
            return step

        # create next step: revise
        self.uacp_agent.db.create_step(
            task_id=step.task_id,
            name=StepTypes.REVISE,
            input=resp["action_parameters"]["content"],
            additional_properties=AdditionalProperties(
                current_plan=self.current_plan.dict(),
                past_steps=str(resp),
            ),
        )

        step.output = str(resp)
        return step

    def revise(self, step: uacp.Step) -> uacp.Step:
        """Review the plan and update the plan accordingly. This behavior will decide if
         the plan is complete or not, and if not, will update the plan accordingly.

        Args:
            step(uacp.Step): The current step.

        Returns:
            uacp.Step: The updated revise step.
        """
        logger.info("[Assistant Agent] Reviewing now.")

        revised_plan: Plan = operations.revise(
            llm=self.llm,
            user_target=self.current_task.input,
            original_plan=self.current_plan.json(),
            past_steps=step.additional_properties["past_steps"],
        )

        Hook.call_hook(
            HookTable.ON_AGENT_REVISE_PLAN, self, revised_plan=revised_plan.json()
        )

        if revised_plan.get_next_task() is None:
            step.is_last = True
            step.output = step.input
            return step
        else:
            # create next step: execute
            task = self.uacp_agent.db.get_task(step.task_id)
            self.uacp_agent.db.create_step(
                task_id=task.task_id,
                name=StepTypes.EXECUTE,
                input=revised_plan.get_next_task().description,
                additional_properties=AdditionalProperties(
                    current_plan=revised_plan.dict()
                ),
            )

            step.output = revised_plan.json()

        return step

    def task_handler(self, task: uacp.Task) -> None:
        """Invoke the task and start planning.

        Args:
            task: Task object
        """
        if not task.input:
            raise Exception("No task prompt")

        logger.info(
            f"[Assistant Agent] Task received. Creating Plan step, task input: {task.input}"  # noqa
        )
        self.current_task_id = task.task_id
        self.uacp_agent.db.create_step(
            task_id=task.task_id, name=StepTypes.PLAN, input=task.input
        )

    def step_handler(self, step: uacp.Step) -> uacp.Step:
        logger.info(
            f"[Assistant Agent] Step received. Executing step, step name: {step.name}"
        )
        step_map: Dict[str, Callable] = {
            StepTypes.PLAN: self.plan,
            StepTypes.EXECUTE: self.execute,
            StepTypes.REVISE: self.revise,
        }

        if len(self.current_task.steps) > self.max_iterations:
            final_output: str = self.current_task.steps[-1].output
            step.output = f"Task has too many steps. Aborting. Recently step output: {final_output}"  # noqa
            return step

        if step.name not in step_map:
            raise ValueError(f"Step name {step.name} not found in step mapping.")

        return step_map[step.name](step)

    def result_handler(self, uacp_task: uacp.Task) -> str:
        return uacp_task.steps[-1].output



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

from typing import Optional

from cushy_storage import CushyOrmCache

from promptulate.error import EmptyMessageSetError
from promptulate.memory.base import BaseChatMemory
from promptulate.pydantic_v1 import Field, validator
from promptulate.schema import MessageSet


class FileChatMemory(BaseChatMemory):
    """Chat message will be stored in the local file cache."""

    file_path: Optional[str] = None
    """If you want to change default store path, you can set specified file_path"""
    cache: CushyOrmCache = Field(default_factory=CushyOrmCache)

    @validator("file_path", always=True)
    def init_cache(cls, file_path: Optional[str]) -> Optional[str]:
        if not file_path:
            return None

        cls.cache = CushyOrmCache(file_path)
        return file_path

    def load_message_set_from_memory(
        self, recently_n: Optional[int] = None
    ) -> MessageSet:
        """Load message from file memory

        Args:
            recently_n: load all messages if it is None, or return recently n messages.

        Returns:
            messages wrapping by MessageSet
        """
        if self.conversation_id not in self.cache:
            raise EmptyMessageSetError()
        return MessageSet.from_listdict_data(self.cache[self.conversation_id])

    def save_message_set_to_memory(self, message_set: MessageSet) -> None:
        self.cache[self.conversation_id] = message_set.listdict_messages



"""Docs: https://docs.litellm.ai/docs/"""
import json
from typing import Optional, TypeVar, Union

import litellm

from promptulate.llms import BaseLLM
from promptulate.pydantic_v1 import BaseModel
from promptulate.schema import (
    AssistantMessage,
    MessageSet,
    StreamIterator,
)
from promptulate.utils.logger import logger

T = TypeVar("T", bound=BaseModel)


def parse_content(chunk) -> (str, str):
    """Parse the litellm chunk.
    Args:
        chunk: litellm chunk.

    Returns:
        content: The content of the chunk.
        ret_data: The additional data of the chunk.
    """
    content = chunk.choices[0].delta.content
    ret_data = json.loads(json.dumps(chunk.json()))
    return content, ret_data


class LiteLLM(BaseLLM):
    def __init__(self, model: str, model_config: Optional[dict] = None, **kwargs):
        logger.info(f"[pne chat] init LiteLLM, model: {model} config: {model_config}")
        super().__init__(**kwargs)
        self._model: str = model
        self._model_config: dict = model_config or {}

    def _predict(
        self, messages: MessageSet, stream: bool = False, *args, **kwargs
    ) -> Union[AssistantMessage, StreamIterator]:
        logger.info(f"[pne chat] prompts: {messages.string_messages}")
        temp_response = litellm.completion(
            model=self._model,
            messages=messages.listdict_messages,
            **self._model_config,
            stream=stream,
        )

        if stream:
            return StreamIterator(
                response_stream=temp_response,
                parse_content=parse_content,
                return_raw_response=False,
            )

        response = AssistantMessage(
            content=temp_response.choices[0].message.content,
            additional_kwargs=temp_response.json()
            if isinstance(temp_response.json(), dict)
            else json.loads(temp_response.json()),
        )
        logger.debug(
            f"[pne chat] response: {json.dumps(response.additional_kwargs, indent=2)}"
        )
        return response

    def __call__(self, instruction: str, *args, **kwargs) -> str:
        return self._predict(
            MessageSet.from_listdict_data(
                [
                    {"content": "You are a helpful assistant.", "role": "system"},
                    {"content": instruction, "role": "user"},
                ]
            )
        ).content



import json
import time
from abc import ABC
from json import JSONDecodeError
from typing import Any, Dict, Iterator, List, Optional, TypeVar, Union

import jwt
import requests

from promptulate.config import pne_config
from promptulate.error import NetWorkError
from promptulate.llms import BaseLLM
from promptulate.preset_roles.prompt import PRESET_SYSTEM_PROMPT
from promptulate.pydantic_v1 import BaseModel
from promptulate.schema import (
    AssistantMessage,
    BaseMessage,
    LLMType,
    MessageSet,
    StreamIterator,
    SystemMessage,
    UserMessage,
)
from promptulate.utils import logger

T = TypeVar("T", bound=BaseModel)


def parse_content(chunk) -> (str, str):
    try:
        ret_data = json.loads(chunk.replace("data: ", ""))
    except JSONDecodeError:
        return None, None
    content: Optional[str] = ret_data["choices"][0]["delta"]["content"]
    return content, ret_data


class ZhiPu(BaseLLM, ABC):
    llm_type: LLMType = LLMType.ZhiPu
    """Used to MessageSet data convert"""
    model: str = "glm-4"
    """Model name to use."""
    enable_private_api_key: bool = False
    """Enable to provide a separate api for openai llm """
    private_api_key: str = ""
    """Store private api key"""
    default_system_prompt: str = ""
    """ernie-bot system message"""
    enable_default_system_prompt: bool = True
    """enable use preset description"""
    exp_seconds: int = 3600
    """set your token expire time"""
    model_config: dict = {}
    """model parameters"""

    @property
    def api_key(self):
        """Select api key to use. private_api_key, openai_api_key, key_pool key is
        optional."""
        if self.enable_private_api_key and self.private_api_key != "":
            return self.private_api_key
        return pne_config.get_zhipuai_api_key()

    def set_private_api_key(self, value: str):
        self.enable_private_api_key = True
        self.private_api_key = value

    def generate_token(self, api_key: str, exp_seconds: int):
        try:
            id, secret = api_key.split(".")
        except ValueError:
            raise ValueError("invalid apikey, please provide a correct key")

        payload = {
            "api_key": id,
            "exp": int(round(time.time() * 1000)) + exp_seconds * 1000,
            "timestamp": int(round(time.time() * 1000)),
        }

        return jwt.encode(
            payload,
            secret,
            algorithm="HS256",
            headers={"alg": "HS256", "sign_type": "SIGN"},
        )

    def __call__(
        self, instruction: str, *args, **kwargs
    ) -> Union[str, BaseMessage, T, List[BaseMessage], StreamIterator]:
        system_message = (
            self.default_system_prompt
            if self.default_system_prompt != ""
            else PRESET_SYSTEM_PROMPT
        )
        if not self.enable_default_system_prompt:
            system_message = ""

        message_set = MessageSet(
            messages=[
                SystemMessage(content=system_message),
                UserMessage(content=instruction),
            ]
        )

        result = self.predict(message_set, **self.model_config)
        if isinstance(result, AssistantMessage):
            return result.content
        else:
            return result

    def _predict(
        self,
        prompts: MessageSet,
        stream: bool = False,
        return_raw_response: bool = False,
        *args,
        **kwargs,
    ) -> Union[str, BaseMessage, T, List[BaseMessage], StreamIterator]:
        """
        Predicts the response using the zhipuai platform.

        Args:
            prompts (MessageSet): The set of prompts to generate a response.
            stream (bool): The set to enable stream.
            return_raw_response (bool): return completion result if true
            **kwargs: zhipu_sdk kwargs

        Returns:
              Return string normally, it means enable_original_return is default False.
              Return BaseMessage if enable_original_return is True.
              Return List[BaseMessage] if stream is True.
              Return T if output_schema is provided.
              Return ZhiPuStreamIterator if stream is enable
        """
        headers = {
            "Content-Type": "application/json",
            "Authorization": self.generate_token(self.api_key, self.exp_seconds),
        }
        body: Dict[str, Any] = self._build_api_params_dict(prompts)
        logger.debug(f"[pne zhipu request] {json.dumps(body, indent=2)}")
        response = requests.post(
            url=pne_config.zhipu_model_url,
            headers=headers,
            stream=stream,
            json=body,
        )
        # return stream
        if stream:
            return StreamIterator(
                response_stream=response.iter_lines(decode_unicode=True),
                parse_content=parse_content,
                return_raw_response=return_raw_response,
            )
        else:
            if response.status_code == 200:
                ret_data = response.json()
                logger.debug(f"[pne zhipu response] {json.dumps(ret_data, indent=2)}")
                content = ret_data["choices"][0]["message"]["content"]
                return AssistantMessage(content=content, additional_kwargs=ret_data)
            else:
                raise NetWorkError(str(response.status_code))

    def _build_api_params_dict(self, prompts: MessageSet) -> Dict[str, Any]:
        dic = {
            "messages": prompts.to_llm_prompt(self.llm_type),
        }
        dic.update({"model": self.model})
        for key in self.model_config:
            if key is not None:
                dic.update({key: self.model_config[key]})
        return dic



from typing import Optional

from promptulate.llms.base import BaseLLM
from promptulate.memory.base import BaseChatMemory, MessageSet
from promptulate.pydantic_v1 import BaseModel
from promptulate.schema import BaseMessage


class BaseMixin(BaseModel):
    llm: BaseLLM
    conversation_id: Optional[str]
    memory: BaseChatMemory

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def embed_message(
        self, cur_message: BaseMessage, message_history: MessageSet
    ) -> None:
        message_history.messages.append(cur_message)
        self.memory.save_message_set_to_memory(message_history)



import re
from typing import Dict, List

from promptulate.llms import BaseLLM, ChatOpenAI
from promptulate.tools import Tool
from promptulate.tools.iot_swith_mqtt.prompt import prompt_template
from promptulate.utils import StringTemplate


class IotSwitchTool(Tool):
    """A tool for switching operations on various devices"""

    name: str = "Iot_Switch_Mqtt"
    description: str = (
        "An IoT tool used for switching operations on various devices."
        "This tool uses the mqtt protocol for transmission."
        "Args: question(str)"
        "The input content is the intention or command to open the specified electrical appliance."  # noqa
        "If the operation of the device is successful, an OK will be returned, otherwise a failure will be returned."  # noqa
    )
    llm_prompt_template: StringTemplate = prompt_template
    rule_table: List[Dict]

    def __init__(
        self, client, llm: BaseLLM = None, rule_table: List[Dict] = None, **kwargs
    ):
        """
        Args:
            llm(BaseLLM): llm, default is ChatOpenAI
            client(paho.mqtt.client.Client): paho mqtt client which has connected.
            rule_table(List[Dict]):
        """
        self.client = client
        self.llm: BaseLLM = llm or ChatOpenAI(
            temperature=0.1, enable_default_system_prompt=False
        )
        self.rule_table = rule_table

        super().__init__(**kwargs)

    def _run(self, question: str, *args, **kwargs) -> str:
        try:
            import paho.mqtt.client as mqtt  # noqa
        except ImportError:
            raise ImportError(
                "Could not import paho python package. "
                "This is needed in order to for IotSwitchTool. "
                "Please install it with `pip install paho-mqtt`."
            )

        if len(self.rule_table) == 0:
            raise ValueError("rule_table is empty")
        else:
            index = 1
            key = ""
            for s in self.rule_table:
                key = key + str(index) + "." + s["content"] + "\n"
                index = index + 1
            prompt = self.llm_prompt_template.format(question=question, rule_key=key)
            llm_output = self.llm(prompt)
            return self._process_llm_result(llm_output)

    def _process_llm_result(self, llm_output: str) -> str:
        answer = re.findall(r"\d+", llm_output)
        if len(answer) == 0:
            return "failure information :" + llm_output
        else:
            topic = self.rule_table[int(answer[0]) - 1]["topic"]
            command = self.rule_table[int(answer[0]) - 1]["ask"]
            self.client.publish(topic, command)
            return "success"



from promptulate.tools.file.toolkit import FileToolKit
from promptulate.tools.file.tools import (
    AppendFileTool,
    CopyFileTool,
    DeleteFileTool,
    ListDirectoryTool,
    MoveFileTool,
    ReadFileTool,
    WriteFileTool,
)

__all__ = [
    "FileToolKit",
    "WriteFileTool",
    "AppendFileTool",
    "ReadFileTool",
    "DeleteFileTool",
    "ListDirectoryTool",
    "CopyFileTool",
    "MoveFileTool",
]



from promptulate.tools.arxiv.toolkit import ArxivTootKit
from promptulate.tools.arxiv.tools import (
    ArxivQueryTool,
    ArxivReferenceTool,
    ArxivSummaryTool,
)

__all__ = [
    "ArxivTootKit",
    "ArxivQueryTool",
    "ArxivReferenceTool",
    "ArxivSummaryTool",
]



from typing import Optional

from typing_extensions import Literal

COLOR_MAPPING = {
    "blue": "36;1",
    "yellow": "33;1",
    "pink": "38;5;200",
    "green": "32;1",
    "red": "31;1",
}
COLOR_TYPE = Literal["blue", "yellow", "pink", "green", "red"]


def print_text(text: str, color: Optional[COLOR_TYPE] = None):
    if not color:
        print(text)
        return

    if color not in COLOR_MAPPING:
        ValueError("Wrong color, only blue, yellow, pink, green, red can be used.")
    print(f"\u001b[{COLOR_MAPPING[color]}m\033[1;3m{text}\u001b[0m")



from abc import ABC, abstractmethod
from typing import Any, Callable, List, Optional

from promptulate.hook import Hook, HookTable
from promptulate.llms import BaseLLM
from promptulate.output_formatter import OutputFormatter
from promptulate.pydantic_v1 import BaseModel


class BaseAgent(ABC):
    """Base class of Agent."""

    def __init__(self, hooks: Optional[List[Callable]] = None, *args, **kwargs):
        hooks = hooks or []
        for hook in hooks:
            Hook.mount_instance_hook(hook, self)

        self._agent_type: str = kwargs.get("agent_type", "Agent")
        self._from = kwargs.get("_from", None)

        Hook.call_hook(HookTable.ON_AGENT_CREATE, self, *args, **kwargs)

    def run(
        self,
        instruction: str,
        output_schema: Optional[type(BaseModel)] = None,
        examples: Optional[List[BaseModel]] = None,
        *args,
        **kwargs,
    ) -> Any:
        """run the tool including specified function and hooks"""
        Hook.call_hook(
            HookTable.ON_AGENT_START,
            self,
            instruction,
            output_schema,
            *args,
            agent_type=self._agent_type,
            **kwargs,
        )

        # get original response from LLM
        result: str = self._run(instruction, *args, **kwargs)

        # Return Pydantic instance if output_schema is specified
        if output_schema:
            formatter = OutputFormatter(output_schema, examples)
            prompt = (
                f"{formatter.get_formatted_instructions()}\n##User input:\n{result}"
            )
            json_response = self.get_llm()(prompt)
            return formatter.formatting_result(json_response)

        Hook.call_hook(
            HookTable.ON_AGENT_RESULT,
            mounted_obj=self,
            result=result,
            agent_type=self._agent_type,
            _from=self._from,
        )
        return result

    @abstractmethod
    def _run(self, instruction: str, *args, **kwargs) -> str:
        """Run the detail agent, implemented by subclass."""
        raise NotImplementedError()

    @abstractmethod
    def get_llm(self) -> BaseLLM:
        """Get the llm when necessary."""
        raise NotImplementedError()



from promptulate.hook import Hook, HookTable
from promptulate.utils.color_print import print_text


class StdOutHook:
    @staticmethod
    def handle_agent_start(*args, **kwargs):
        if kwargs.get("agent_type", None):
            print_text(f"[Agent] {kwargs['agent_type']} start...", "red")
        else:
            print_text("[Agent] Agent Start...", "red")

        if kwargs.get("_from", None) is None:
            print_text(f"[User instruction] {args[0]}", "blue")
        elif kwargs["_from"] == "agent":
            print_text(f"[Agent execution] {args[0]}", "blue")

    @staticmethod
    def handle_agent_plan(*args, **kwargs):
        print_text(f"[Plan] {kwargs['plan']}", "green")

    @staticmethod
    def handle_agent_revise_plan(*args, **kwargs):
        print_text(f"[Revised Plan] {kwargs['revised_plan']}", "green")

    @staticmethod
    def handle_agent_action(*args, **kwargs):
        print_text(f"[Thought] {kwargs['thought']}", "yellow")
        print_text(
            f"[Action] {kwargs['action']} args: {kwargs['action_input']}", "yellow"
        )

    @staticmethod
    def handle_agent_observation(*args, **kwargs):
        print_text(f"[Observation] {kwargs['observation']}", "yellow")

    @staticmethod
    def handle_agent_result(*args, **kwargs):
        if kwargs.get("_from", None) is None:
            print_text(f"[Agent Result] {kwargs['result']}", "green")
            print_text("[Agent] Agent End.", "pink")
        elif kwargs["_from"] == "agent":
            print_text(f"[Execute Result] {kwargs['result']}", "green")
            print_text("[Execute] Execute End.", "pink")

    @staticmethod
    def registry_stdout_hooks():
        """Registry and enable stdout hooks. StdoutHook can print colorful
        information."""
        Hook.registry_hook(
            HookTable.ON_AGENT_START, StdOutHook.handle_agent_start, "component"
        )
        Hook.registry_hook(
            HookTable.ON_AGENT_PLAN, StdOutHook.handle_agent_plan, "component"
        )
        Hook.registry_hook(
            HookTable.ON_AGENT_REVISE_PLAN,
            StdOutHook.handle_agent_revise_plan,
            "component",
        )
        Hook.registry_hook(
            HookTable.ON_AGENT_ACTION, StdOutHook.handle_agent_action, "component"
        )
        Hook.registry_hook(
            HookTable.ON_AGENT_OBSERVATION,
            StdOutHook.handle_agent_observation,
            "component",
        )
        Hook.registry_hook(
            HookTable.ON_AGENT_RESULT, StdOutHook.handle_agent_result, "component"
        )

    @staticmethod
    def unregister_stdout_hooks():
        Hook.unregister_hook(StdOutHook.handle_agent_start)
        Hook.unregister_hook(StdOutHook.handle_agent_action)
        Hook.unregister_hook(StdOutHook.handle_agent_observation)
        Hook.unregister_hook(StdOutHook.handle_agent_result)



from enum import Enum
from typing import List, Optional

from promptulate.pydantic_v1 import BaseModel, Field


class TaskStatus(str, Enum):
    TODO = "todo"
    DONE = "done"
    ERROR = "error"
    DISCARDED = "discarded"


class Task(BaseModel):
    task_id: int = Field(..., description="The ID of the task. Start from 1.")
    description: str = Field(..., description="The description of the task.")
    status: TaskStatus = Field(TaskStatus.TODO, description="The status of the task.")


class AgentPlanResponse(BaseModel):
    goals: List[str] = Field(..., description="List of goals in the plan.")
    tasks: List[Task] = Field(
        ..., description="List of tasks in the plan, should be in sorted order"
    )


class Plan(AgentPlanResponse):
    next_task_id: Optional[int] = Field(
        ..., description="The ID of the next task, null if no more tasks are needed."
    )

    def get_next_task(self) -> Optional[Task]:
        if self.next_task_id is None:
            return None
        return next((t for t in self.tasks if t.task_id == self.next_task_id), None)


class AgentReviseResponse(BaseModel):
    thought: str = Field(..., description="The thought of the reflect plan.")
    goals: List[str] = Field(..., description="List of goals in the plan.")
    tasks: List[Task] = Field(
        ..., description="List of tasks in the plan, should be in sorted order"
    )
    next_task_id: Optional[int] = Field(
        ..., description="The ID of the next task, null if no more tasks are needed."
    )


plan_examples = [
    AgentPlanResponse(
        goals=["Goal 1"],
        tasks=[
            Task(task_id=1, description="Task 1", status=TaskStatus.TODO),
            Task(task_id=2, description="Task 2", status=TaskStatus.TODO),
        ],
    ),
]


revise_plan_examples = [
    AgentReviseResponse(
        thought="thought what to do next",
        goals=["Goal 1"],
        tasks=[
            Task(task_id=1, description="Task 1", status=TaskStatus.DONE),
            Task(task_id=2, description="Task 2", status=TaskStatus.TODO),
        ],
        next_task_id=2,
    ),
]



import os
from typing import Any, Callable, Optional

from promptulate.uacp.db import InMemoryTaskDB, TaskDB
from promptulate.uacp.schema import Status, Step, Task
from promptulate.utils.logger import logger

StepHandler = Callable[[Step], Step]
TaskHandler = Callable[[Task], None]
ResultHandler = Callable[[Task], Any]

_task_handler: Optional[TaskHandler]
_step_handler: Optional[StepHandler]


class Agent:
    """Unified Agent Communication Protocol (UACP) agent. You can implement the
    `TaskHandler`, `StepHandler`, and `ResultHandler` to customize the agent's
    behavior.

    TaskHandler: A function that handles the task, task handler will be called when
    the task is created.

    StepHandler: A function that handles the step, step handler will be called when
    the step is created.

    ResultHandler: A function that handles the result, result handler will be called
    when the task is completed.
    """

    def __init__(
        self,
        task_handler: TaskHandler,
        step_handler: StepHandler,
        result_handler: ResultHandler,
        *,
        db: TaskDB = None,
        workspace: str = None,
    ):
        self.task_handler: TaskHandler = task_handler
        self.step_handler: StepHandler = step_handler
        self.result_handler: ResultHandler = result_handler
        self.db: TaskDB = db or InMemoryTaskDB()
        self.workspace: str = workspace or os.getenv("AGENT_WORKSPACE", "workspace")

    def run(
        self,
        input: Optional[str] = None,
        additional_input: Optional[dict] = None,
    ) -> Any:
        """Run the agent with the specified input and additional input.

        Args:
            input(Optional[str]): The input for the agent.
            additional_input: Additional input for the agent.

        Returns:
            Any: The result of the agent.
        """
        task = self.db.create_task(input=input, additional_input=additional_input)

        # task handler to create the first step
        self.task_handler(task)

        # step handler to execute the steps
        task = self.db.get_task(task.task_id)

        if not task.steps:
            raise Exception("No steps to execute")

        while True:
            task = self.db.get_task(task.task_id)
            step: Optional[Step] = next(
                filter(lambda x: x.status == Status.created, task.steps), None
            )

            if step is None:
                break

            step.status = Status.running
            step = self.step_handler(step)
            step.status = Status.completed

            self.db.update_step(task.task_id, step)
            logger.info(f"[uacp] Finish step, name: {step.name} data: {step.json()}")

            if step.is_last:
                break

        # result handler to process the task result
        task = self.db.get_task(task.task_id)
        task_result: Any = self.result_handler(task)
        self.db.update_task(task)

        return task_result

    def startup_server(self):
        pass
